package cs108.puzzle;

import static javafx.beans.binding.Bindings.when;

import java.util.EnumMap;
import java.util.Iterator;
import java.util.Map;

import javafx.application.Application;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.geometry.VPos;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Text;
import javafx.stage.Stage;

public final class Main extends Application {
    private final static int PIECE_SIDE = 50;
    private final static int INTER_PIECE_SPACE = 3;

    public static void main(String[] args) {
        launch(args);
    }

    @Override
    public void start(Stage primaryStage) throws Exception {
        Pane gamePane = new Pane();
        double prefSide = PIECE_SIDE * Puzzle.SIDE_LENGTH + INTER_PIECE_SPACE * (Puzzle.SIDE_LENGTH + 1);
        gamePane.setPrefSize(prefSide, prefSide);

        Map<Piece, Node> pieceRepr = new EnumMap<>(Piece.class);
        for (Piece p: Piece.values()) {
            if (p == Piece.HOLE)
                continue;
            pieceRepr.put(p, nodeForPiece(p));
        }

        gamePane.getChildren().setAll(pieceRepr.values());

        Puzzle puzzle = new Puzzle();
        layoutPuzzle(puzzle, pieceRepr);
        puzzle.addObserver(p -> layoutPuzzle(puzzle, pieceRepr));

        pieceRepr.forEach((p, n) -> {
            n.setOnMouseClicked(e -> {
                int pIndex = puzzle.pieces().indexOf(p);
                if (puzzle.canSwap(pIndex))
                    puzzle.swap(pIndex);
                else
                    java.awt.Toolkit.getDefaultToolkit().beep();
            });
        });

        Button shuffleButton = new Button("Mélanger !");
        shuffleButton.setOnAction(e -> puzzle.shuffle());

        BooleanProperty solved = new SimpleBooleanProperty(puzzle.isSolved());
        puzzle.addObserver(p -> solved.set(puzzle.isSolved()));

        Text motivator = new Text();
        motivator.textProperty().bind(when(solved).then("Bravo!").otherwise("à résoudre..."));

        BorderPane mainPane = new BorderPane(gamePane, motivator, null, shuffleButton, null);

        primaryStage.setScene(new Scene(mainPane));
        primaryStage.setTitle("Taquin");
        primaryStage.show();
    }

    private Node nodeForPiece(Piece p) {
        assert p != Piece.HOLE;

        Rectangle r = new Rectangle(PIECE_SIDE, PIECE_SIDE);
        r.setFill(Color.LIGHTSALMON);
        Text t = new Text(5, 5, p.repr());
        t.setTextOrigin(VPos.TOP);
        return new Group(r, t);
    }

    private void layoutPuzzle(Puzzle puzzle, Map<Piece, Node> pieceRepr) {
        Iterator<Piece> pIt = puzzle.pieces().iterator();
        for (int r = 0; r < Puzzle.SIDE_LENGTH; ++r) {
            int y = INTER_PIECE_SPACE + r * (PIECE_SIDE + INTER_PIECE_SPACE);
            for (int c = 0; c < Puzzle.SIDE_LENGTH; ++c) {
                Node n = pieceRepr.get(pIt.next());
                if (n == null)
                    continue;

                int x = INTER_PIECE_SPACE + c * (PIECE_SIDE + INTER_PIECE_SPACE);
                n.relocate(x, y);
            }
        }
    }
}
