package cs108.puzzle;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public final class Puzzle extends AbstractSubject {
    public static final int SIDE_LENGTH = 3;
    public static final List<Piece> SOLVED =
            Collections.unmodifiableList(Arrays.asList(Piece.values()));

    public static int posToIndex(int x, int y) {
        assert isValidPos(x, y);
        return y * SIDE_LENGTH + x;
    }

    private final List<Piece> pieces = new ArrayList<>(SOLVED);

    public List<Piece> pieces() {
        return Collections.unmodifiableList(pieces);
    }

    public void shuffle() {
        Collections.shuffle(pieces);
        notifyObservers();
    }

    public boolean isSolved() {
        return pieces.equals(SOLVED);
    }

    public boolean canSwap(int index) {
        return indexOfNeighboringHole(index) != -1;
    }

    public void swap(int index) {
        assert canSwap(index);
        Collections.swap(pieces, index, indexOfNeighboringHole(index));
        notifyObservers();
    }

    private static boolean isValidPos(int x, int y) {
        return (0 <= x && x < SIDE_LENGTH)
                && (0 <= y && y < SIDE_LENGTH);
    }

    private boolean isHoleAt(int x, int y) {
        return isValidPos(x, y) && pieces.get(posToIndex(x, y)) == Piece.HOLE;
    }

    private int indexOfNeighboringHole(int index) {
        int x = index % SIDE_LENGTH;
        int y = index / SIDE_LENGTH;

        if (isHoleAt(x - 1, y))
            return posToIndex(x - 1, y);
        if (isHoleAt(x + 1, y))
            return posToIndex(x + 1, y);
        if (isHoleAt(x, y - 1))
            return posToIndex(x, y - 1);
        if (isHoleAt(x, y + 1))
            return posToIndex(x, y + 1);
        return -1;
    }
}
